/*
PeRdr - PE file disassembler
Copyright (C) 1999-2006 Frediano Ziglio
-----

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-----

INFORMATION
  www: http://freddy77.tripod.com/perdr/
  e-mail: freddy77@angelfire.com
*/
#include "global.h"
#pragma hdrstop

#include <stdexcept>
#include "x86istr.h"
#include "x86dasm.h"
#ifdef DEBUG
#include <cctype>
#endif
#include "utils/f77auto_ptr"

using namespace std;

void WriteInstruction(const Instruction& instruction)
{
  char buffer[80];
  instruction.Write(buffer);
  printf("%s\n",buffer);
}

class DataReaderTest: public DataReader
{
  vma_t start;
  const unsigned char* data;
  void ReadNext() { throw OutOfAddress(start + (ptr-data)); }
public:
  DataReaderTest(vma_t address,const unsigned char* code,size_t len):
    DataReader(code,code+len),start(address),data(code)
  {}
  void UnReadU8() { --ptr; }
  vma_t Tell() { return start + (ptr - data); }
};

int GetInstruction(vma_t address,const unsigned char* code,size_t len,Instruction& instruction)
{
  // legge l'istruzione
  RelocationInfos rels;
  InstructionDecoder decoder(rels);
  DataReaderTest reader(address, code, len);
  return decoder.Decode(instruction,reader);
}

static int hexdigit(int c)
{
	if (c >= '0' && c <= '9')
		return c - '0';
	/* ascii optimization, 'A' -> 'a', 'a' -> 'a' */
	c |= 0x20;
	if (c >= 'a' && c <= 'f')
		return c - 'a' + 10;
	return 0;
}

int main(int argc, char* argv[])
{
  Instruction instr;

  if (argc == 1)
  {
    if (GetInstruction(0x401000,(unsigned char*)"pippo",5,instr))
      WriteInstruction(instr);

    if (GetInstruction(0x401000,(unsigned char*)"\x0f\xb6\x40\x83",4,instr))
      WriteInstruction(instr);

    if (GetInstruction(0x401000,(unsigned char*)"\x0f\xb7\x40\x83",4,instr))
      WriteInstruction(instr);

    if (GetInstruction(0x401000,(unsigned char*)"\x0f\xbe\xc1",3,instr))
      WriteInstruction(instr);

    return 0;
  }

  if (strcmp(argv[1],"/?") == 0 || strcmp(argv[1],"--help") == 0)
  {
    printf("x86test [address] -x bytes | file [begin [-length|end]]\n"
           "\n"
           "address - base address of disassembly (default is 401000)\n"
           "bytes   - bytes to disassemble\n"
           "file    - file to disassemble\n"
           "begin   - starting position within file (default is 0)\n"
           "length  - number of instructions to disassemble (default is 10 if begin)\n"
           "end     - ending position to stop disassembly (whole file if not begin)\n"
           "          if end is smaller than begin, end becomes length\n"
           "\n"
           "address & bytes are hexadecimal; prefix begin/length/end with '0x' for hex.\n"
           );
    return 0;
  }

  vma_t pc;
  char* end;
  int  arg;
  pc = strtoul(argv[1],&end,16);
  if (pc == 0 || *end)
  {
    pc = 0x401000;
    arg = 1;
  }
  else
    arg = 2;

  unsigned char* buf;
  int blen, dlen = 0;
  int ic = 0;
  if (strcmp(argv[arg],"-x") == 0)
  {
    ++arg;
    int len = 0;
    for (int j = arg; j < argc; ++j)
      len += (strlen(argv[j])+1) / 2;
    if (len == 0)
      return 0;
    buf = new unsigned char[len];
    dlen = 0;
    for (int j = arg; j < argc; ++j)
      for (int k = 0; argv[j][k]; k += 2)
      {
        buf[dlen] = hexdigit(argv[j][k]) * 16 + hexdigit(argv[j][k+1]);
        ++dlen;
        if (argv[j][k+1] == '\0')
          break;
      }
    blen = dlen;
  }
  else
  {
    FILE* file = fopen(argv[arg],"rb");
    if (file == NULL)
    {
      printf("Cannot open %s\n", argv[arg]);
      return 1;
    }
    if (arg+1 == argc)
    {
      fseek(file,0,SEEK_END);
      dlen = ftell(file);
      rewind(file);
    }
    else
    {
      ic = strtol(argv[arg+1],0,0);
      if (ic < 0)
        ic = -ic;
      else
      {
        fseek(file,ic,SEEK_SET);
        if (arg+2 == argc)
          ic = 10;
        else
        {
          ic = strtol(argv[arg+2],0,0);
          if (ic < 0)
            ic = -ic;
          else if (ic >= ftell(file))
          {
            dlen = ic - ftell(file);
            ic = 0;
          }
        }
      }
    }
    if (ic)
      dlen = ic * 15;
    blen = dlen + 15;
    buf = new unsigned char[blen];
    blen = fread(buf,1,blen,file);
    if (blen < dlen)
      dlen = blen;
    fclose(file);
  }
  int len;
  int ofs = 0;
  do
  {
    len = GetInstruction(pc+ofs,buf+ofs,(blen>15)?15:blen,instr);
    printf("%08X ",pc+ofs);
    if (len == 0)
    {
      printf("%02X failed\n",buf[ofs]);
      break;
    }
    for (int j = 0; j < len; ++j)
      printf("%02X",buf[ofs+j]);
    for (int j = len; j < 11; ++j)
      printf("  ");
    printf("\t");
    WriteInstruction(instr);
    ofs += len;
    blen -= len;
    dlen -= len;
  } while (dlen > 0 && --ic != 0);

  delete[] buf;

  return 0;
}
